"""This module contains security related CLI tools and code"""
import json
import re
import subprocess
import sys
from dataclasses import (
    asdict,
    dataclass,
)
from enum import Enum
from functools import partial
from inspect import cleandoc
from typing import (
    Generator,
    Iterable,
    Tuple,
)

import typer

stdout = print
stderr = partial(print, file=sys.stderr)


# Note: In the long term we may want to adapt the official CVE json schema
# https://github.com/CVEProject/cve-schema/blob/master/schema/v5.0/CVE_JSON_5.0_schema.json
@dataclass(frozen=True)
class Issue:
    cve: str
    cwe: str
    description: str
    coordinates: str
    references: tuple


def _issues(input) -> Generator[Issue, None, None]:
    lines = (l for l in input if l.strip() != "")
    for line in lines:
        obj = json.loads(line)
        yield Issue(**obj)


def _issues_as_json_str(issues):
    for issue in issues:
        issue = asdict(issue)
        yield json.dumps(issue)


def gh_security_issues() -> Generator[Tuple[str, str], None, None]:
    """
    Yields issue-id, cve-id pairs for all (closed, open) issues associated with CVEs

    Return:
        A generator which yields tuples of (id,title).

    Raises:
        subprocess.CalledProcessError: If the underlying command fails.
    """
    # fmt: off
    command = [
        "gh", "issue", "list",
        "--label", "security",
        "--search", "CVE",
        "--json", "id,title",
        "--limit", "1000",
    ]
    # fmt: on
    try:
        result = subprocess.run(command, check=True, capture_output=True)
    except FileNotFoundError as ex:
        msg = "Command 'gh' not found. Please make sure you have installed the github cli."
        raise FileNotFoundError(msg) from ex
    except subprocess.CalledProcessError as ex:
        stderr(f"{ex}")
        raise ex

    cve_pattern = re.compile(r"CVE-\d{4}-\d{4,7}")
    issues = json.loads(result.stdout.decode("utf-8"))
    issues = (
        (issue["id"], cve_pattern.search(issue["title"]).group())  # type: ignore
        for issue in issues
        if cve_pattern.search(issue["title"])
    )
    return issues


def from_maven(report: str) -> Iterable[Issue]:
    # Note: Consider adding warnings if there is the same cve with multiple coordinates
    report = json.loads(report)
    dependencies = report["vulnerable"]  # type: ignore
    for _, dependency in dependencies.items():  # type: ignore
        for v in dependency["vulnerabilities"]:  # type: ignore
            references = [v["reference"]] + v["externalReferences"]
            yield Issue(
                cve=v["cve"],
                cwe=v["cwe"],
                description=v["description"],
                coordinates=dependency["coordinates"],
                references=tuple(references),
            )


def security_issue_title(issue: Issue) -> str:
    return f"🔐 {issue.cve}: {issue.coordinates}"


def security_issue_body(issue: Issue) -> str:
    def as_markdown_listing(elements: Iterable[str]):
        return "\n".join(f"- {element}" for element in elements)

    body = cleandoc(
        """
        ## Summary
        {description}

        CVE: {cve}
        CWE: {cwe}

        ## References
        {references}
        """
    )
    return body.format(
        cve=issue.cve,
        cwe=issue.cwe,
        description=issue.description,
        references=as_markdown_listing(issue.references),
    )


def create_security_issue(issue: Issue) -> Tuple[str, str]:
    # fmt: off
    command = [
        "gh", "issue", "create",
        "--label", "security",
        "--title", security_issue_title(issue),
        "--body", security_issue_body(issue),
    ]
    # fmt: on
    try:
        result = subprocess.run(command, check=True, capture_output=True)
    except FileNotFoundError as ex:
        msg = "Command 'gh' not found. Please make sure you have installed the github cli."
        raise FileNotFoundError(msg) from ex
    except subprocess.CalledProcessError as ex:
        stderr(f"{ex}")
        raise ex

    std_err = result.stderr.decode("utf-8")
    std_out = result.stdout.decode("utf-8")

    return std_err, std_out


CLI = typer.Typer()
CVE_CLI = typer.Typer()
CLI.add_typer(CVE_CLI, name="cve", help="Work with CVE's")


class Format(str, Enum):
    Maven = "maven"


# pylint: disable=redefined-builtin
@CVE_CLI.command(name="convert")
def convert(
    format: Format = typer.Argument(..., help="input format to be converted."),
    input_file: typer.FileText = typer.Argument(
        default="-", mode="r", help="input file which shall be converted"
    ),
) -> None:
    """
    Convert a language or tool specific security report into a list of cve's in the jsonl format

    Output:
    { "cve": "<cve-id>", "cwe": "<cwe-id>", "description": "<multiline string>", "coordinates": "<string>", "references": ["<url>", "<url>", ...] }
    """

    def _maven(infile):
        issues = from_maven(infile.read())
        for issue in _issues_as_json_str(issues):
            stdout(issue)
        raise typer.Exit(code=0)

    actions = {Format.Maven: _maven}
    action = actions[format]
    action(input_file)


class Filter(str, Enum):
    GitHubIssues = "github-issues"
    PassThrough = "pass-through"


# pylint: disable=redefined-builtin
@CVE_CLI.command(name="filter")
def filter(
    type: Filter = typer.Argument(help="filter type to apply"),
    input_file: typer.FileText = typer.Argument(
        default="-", mode="r", help="file containing cve's in the jsonl format"
    ),
) -> None:
    """
    Filter specific CVE's from the input


    Input:
    { "cve": "<cve-id>", "cwe": "<cwe-id>", "description": "<multiline string>", "coordinates": "<string>", "references": ["<url>", "<url>", ...] }
    """

    def _github(infile):
        to_be_filtered = {cve for _, cve in gh_security_issues()}
        stderr("Filtering:")
        for issue in to_be_filtered:
            stderr(f"{issue}")
        filtered_issues = [
            issue for issue in _issues(infile) if issue.cve not in to_be_filtered
        ]
        for issue in _issues_as_json_str(filtered_issues):
            stdout(issue)
        raise typer.Exit(code=0)

    def _pass_through(infile):
        for line in infile:
            stdout(line)
        raise typer.Exit(code=0)

    actions = {Filter.GitHubIssues: _github, Filter.PassThrough: _pass_through}
    action = actions[type]
    action(input_file)


@CVE_CLI.command(name="create")
def create(
    input_file: typer.FileText = typer.Argument(
        default="-", mode="r", help="file of cve's in the jsonl format"
    ),
) -> None:
    """
    Create GitHub issues for CVE's

    Input:
    { "cve": "<cve-id>", "cwe": "<cwe-id>", "description": "<multiline string>", "coordinates": "<string>", "references": ["<url>", "<url>", ...] }

    Output:
    Links to the created issue(s)
    """
    for issue in _issues(input_file):
        std_err, std_out = create_security_issue(issue)
        stderr(std_err)
        stdout(std_out)


if __name__ == "__main__":
    CLI()



# ATTENTION:
# This file is generated by exasol/toolbox/pre_commit_hooks/package_version.py when using:
#   * either "poetry run nox -s fix"
#   * or "poetry run version-check <path/version.py> --fix"
# Do not edit this file manually!
# If you need to change the version, do so in the project.toml, e.g. by using `poetry version X.Y.Z`.
MAJOR = 0
MINOR = 6
PATCH = 1
VERSION = f"{MAJOR}.{MINOR}.{PATCH}"



import argparse
import contextlib
import glob
import itertools
import json
import logging
import multiprocessing
import os
import pathlib
import re
import shutil
import string
import subprocess
import sys
import tempfile
from importlib import resources

from jinja2 import Template
from rich.console import Console
from rich.logging import RichHandler
from sphinx import config as sphinx_config
from sphinx import project as sphinx_project

from exasol.toolbox.sphinx.multiversion import (
    git,
    sphinx,
)

logging.basicConfig(
    level="INFO",
    datefmt="[%X]",
    handlers=[RichHandler(console=Console(stderr=True), show_time=False)],
)
logger = logging.getLogger(__name__)


@contextlib.contextmanager
def working_dir(path):
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def load_sphinx_config_worker(q, confpath, confoverrides, add_defaults):
    try:
        with working_dir(confpath):
            current_config = sphinx_config.Config.read(
                confpath,
                confoverrides,
            )

        if add_defaults:
            current_config.add(
                "smv_tag_whitelist", sphinx.DEFAULT_TAG_WHITELIST, "html", str
            )
            current_config.add(
                "smv_branch_whitelist",
                sphinx.DEFAULT_TAG_WHITELIST,
                "html",
                str,
            )
            current_config.add(
                "smv_remote_whitelist",
                sphinx.DEFAULT_REMOTE_WHITELIST,
                "html",
                str,
            )
            current_config.add(
                "smv_released_pattern",
                sphinx.DEFAULT_RELEASED_PATTERN,
                "html",
                str,
            )
            current_config.add(
                "smv_outputdir_format",
                sphinx.DEFAULT_OUTPUTDIR_FORMAT,
                "html",
                str,
            )
            current_config.add(
                "smv_build_targets",
                sphinx.DEFAULT_BUILD_TARGETS,
                "html",
                {str: {str: str}},
            )
            current_config.add("smv_prefer_remote_refs", False, "html", bool)
            current_config.add("smv_prebuild_command", "", "html", str)
            current_config.add("smv_prebuild_export_pattern", "", "html", str)
            current_config.add("smv_prebuild_export_destination", "", "html", str)
            current_config.add("smv_postbuild_command", "", "html", str)
            current_config.add("smv_postbuild_export_pattern", "", "html", str)
            current_config.add("smv_postbuild_export_destination", "", "html", str)
            current_config.add(
                "smv_clean_intermediate_files",
                sphinx.DEFAULT_CLEAN_INTERMEDIATE_FILES_FLAG,
                "html",
                bool,
            )
        current_config.pre_init_values()
        current_config.init_values()
    except Exception as err:
        q.put(err)
        return

    q.put(current_config)


def load_sphinx_config(confpath, confoverrides, add_defaults=False):
    q = multiprocessing.Queue()
    proc = multiprocessing.Process(
        target=load_sphinx_config_worker,
        args=(q, confpath, confoverrides, add_defaults),
    )
    proc.start()
    proc.join()
    result = q.get_nowait()
    if isinstance(result, Exception):
        raise result
    return result


def get_python_flags():
    if sys.flags.bytes_warning:
        yield "-b"
    if sys.flags.debug:
        yield "-d"
    if sys.flags.hash_randomization:
        yield "-R"
    if sys.flags.ignore_environment:
        yield "-E"
    if sys.flags.inspect:
        yield "-i"
    if sys.flags.isolated:
        yield "-I"
    if sys.flags.no_site:
        yield "-S"
    if sys.flags.no_user_site:
        yield "-s"
    if sys.flags.optimize:
        yield "-O"
    if sys.flags.quiet:
        yield "-q"
    if sys.flags.verbose:
        yield "-v"
    for option, value in sys._xoptions.items():
        if value is True:
            yield from ("-X", option)
        else:
            yield from ("-X", f"{option}={value}")


def find_matching_files_and_dirs(pattern: str, root_dir: str) -> set:
    logger = logging.getLogger(__name__)
    logger.debug(f"Export pattern: {pattern}")
    regex = re.compile(pattern)
    matches = set()
    for root, dirs, files in os.walk(root_dir):
        for directory in dirs:
            directory_result = regex.match(directory)
            if directory_result:
                matches.add(os.path.join(root, directory))
        for file in files:
            file_result = regex.match(file)
            if file_result:
                matches.add(os.path.join(root, file))
    logger.debug(f"Pattern matches: {matches}")

    return matches


def export_to_destination(items: set, destination: str) -> None:
    logger = logging.getLogger(__name__)
    os.makedirs(destination, exist_ok=True)
    for item in items:
        # Copy file to desination
        if os.path.isfile(item):
            logger.debug(
                "Copying {} to {}".format(
                    item, os.path.join(destination, os.path.basename(item))
                )
            )
            shutil.copy(item, os.path.join(destination, os.path.basename(item)))

        # Copy directory to desination
        elif os.path.isdir(item):
            logger.debug(f"Copying {item} to {destination}")
            shutil.copytree(item, destination, dirs_exist_ok=True)


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("sourcedir", help="path to documentation source files")
    parser.add_argument("outputdir", help="path to output directory")
    parser.add_argument(
        "filenames",
        nargs="*",
        help="a list of specific files to rebuild. Ignored if -a is specified",
    )
    parser.add_argument(
        "-c",
        metavar="PATH",
        dest="confdir",
        help=(
            "path where configuration file (conf.py) is located "
            "(default: same as SOURCEDIR)"
        ),
    )
    parser.add_argument(
        "-C",
        action="store_true",
        dest="noconfig",
        help="use no config file at all, only -D options",
    )
    parser.add_argument(
        "-D",
        metavar="setting=value",
        action="append",
        dest="define",
        default=[],
        help="override a setting in configuration file",
    )
    parser.add_argument(
        "--dump-metadata",
        action="store_true",
        help="dump generated metadata and exit",
    )
    return parser


def main(argv=None):
    parser = _create_parser()
    if not argv:
        argv = sys.argv[1:]
    args, argv = parser.parse_known_args(argv)

    try:
        if args.noconfig:
            sys.exit(1)
        sys.exit(_main(args, argv))
    except Exception as ex:
        logger.exception(ex)
        sys.exit(1)


def _main(args, argv):
    sourcedir_absolute = os.path.abspath(args.sourcedir)
    confdir_absolute = (
        os.path.abspath(args.confdir)
        if args.confdir is not None
        else sourcedir_absolute
    )

    # Conf-overrides
    confoverrides = {}
    for d in args.define:
        key, _, value = d.partition("=")
        confoverrides[key] = value

    # Parse config
    config = load_sphinx_config(confdir_absolute, confoverrides, add_defaults=True)

    # Get relative paths to root of git repository
    gitroot = pathlib.Path(git.get_toplevel_path(cwd=sourcedir_absolute)).resolve()
    cwd_absolute = os.path.abspath("")
    cwd_relative = os.path.relpath(cwd_absolute, str(gitroot))

    logger.debug("Git toplevel path: %s", str(gitroot))
    sourcedir = os.path.relpath(sourcedir_absolute, str(gitroot))
    logger.debug("Source dir (relative to git toplevel path): %s", str(sourcedir))
    if args.confdir:
        confdir = os.path.relpath(confdir_absolute, str(gitroot))
    else:
        confdir = sourcedir
    logger.debug("Conf dir (relative to git toplevel path): %s", str(confdir))
    conffile = os.path.join(confdir, "conf.py")

    # Get git references
    gitrefs = git.get_refs(
        str(gitroot),
        config.smv_tag_whitelist,
        config.smv_branch_whitelist,
        config.smv_remote_whitelist,
        files=(sourcedir, conffile),
    )

    # Order git refs
    if config.smv_prefer_remote_refs:
        gitrefs = sorted(gitrefs, key=lambda x: (not x.is_remote, *x))
    else:
        gitrefs = sorted(gitrefs, key=lambda x: (x.is_remote, *x))

    with tempfile.TemporaryDirectory() as tmp:
        # Generate Metadata
        metadata = {}
        outputdirs = set()
        for gitref in gitrefs:
            # Clone Git repo
            repopath = os.path.join(tmp, gitref.commit)
            try:
                git.copy_tree(str(gitroot), gitroot.as_uri(), repopath, gitref)
            except (OSError, subprocess.CalledProcessError):
                logger.error(
                    "Failed to copy git tree for %s to %s",
                    gitref.refname,
                    repopath,
                )
                continue

            # Find config
            confpath = os.path.join(repopath, confdir)
            try:
                current_config = load_sphinx_config(confpath, confoverrides)
            except (OSError, sphinx_config.ConfigError):
                logger.error(
                    "Failed load config for %s from %s",
                    gitref.refname,
                    confpath,
                )
                continue

            # Ensure that there are not duplicate output dirs
            outputdir = config.smv_outputdir_format.format(
                ref=gitref,
                config=current_config,
            )
            if outputdir in outputdirs:
                logger.warning(
                    "outputdir '%s' for %s conflicts with other versions",
                    outputdir,
                    gitref.refname,
                )
                continue
            outputdirs.add(outputdir)

            # Get List of files
            source_suffixes = current_config.source_suffix
            if isinstance(source_suffixes, str):
                source_suffixes = [current_config.source_suffix]

            current_sourcedir = os.path.join(repopath, sourcedir)
            project = sphinx_project.Project(current_sourcedir, source_suffixes)
            metadata[gitref.name] = {
                "name": gitref.name,
                "project": current_config.project,
                "version": current_config.version,
                "release": current_config.release,
                "rst_prolog": current_config.rst_prolog,
                "is_released": bool(
                    re.match(config.smv_released_pattern, gitref.refname)
                ),
                "source": gitref.source,
                "creatordate": gitref.creatordate.strftime(sphinx.DATE_FMT),
                "basedir": repopath,
                "sourcedir": current_sourcedir,
                "outputdir": os.path.join(os.path.abspath(args.outputdir), outputdir),
                "confdir": confpath,
                "docnames": list(project.discover()),
                "build_targets": config.smv_build_targets,
            }

        if args.dump_metadata:
            print(json.dumps(metadata, indent=2))
            return 0

        if not metadata:
            logger.error("No matching refs found!")
            return 2

        # Write Metadata
        metadata_path = os.path.abspath(os.path.join(tmp, "versions.json"))
        with open(metadata_path, mode="w") as fp:
            json.dump(metadata, fp, indent=2)

        # Run Sphinx
        argv.extend(["-D", f"smv_metadata_path={metadata_path}"])
        for version_name, data in metadata.items():
            os.makedirs(data["outputdir"], exist_ok=True)

            defines = itertools.chain(
                *(("-D", string.Template(d).safe_substitute(data)) for d in args.define)
            )

            current_argv = argv.copy()
            current_argv.extend(
                [
                    *defines,
                    "-D",
                    f"smv_current_version={version_name}",
                    "-c",
                    confdir_absolute,
                    *args.filenames,
                ]
            )
            current_cwd = os.path.join(data["basedir"], cwd_relative)
            if config.smv_prebuild_command != "":
                logger.debug(
                    "Running prebuild command: %r", config.smv_prebuild_command
                )
                subprocess.check_call(
                    config.smv_prebuild_command, cwd=current_cwd, shell=True
                )

                if config.smv_prebuild_export_pattern != "":
                    matches = find_matching_files_and_dirs(
                        config.smv_prebuild_export_pattern, current_cwd
                    )
                    export_dst = os.path.join(
                        data["outputdir"],
                        config.smv_prebuild_export_destination,
                    )
                    export_to_destination(matches, export_dst)

            for build_target_name, build_target in data["build_targets"].items():
                builder = build_target.get("builder", None)
                if not builder:
                    raise AttributeError(
                        "Builder for build target {} not defined".format(
                            build_target_name
                        )
                    )

                downloadable = build_target.get("downloadable", False)
                download_format = build_target.get("download_format", "")

                if downloadable and download_format == "":
                    raise ValueError(
                        (
                            "Download format for build target {} not defined "
                            "but downloadable is True"
                        ).format(build_target_name)
                    )

                flag = "-b"
                builder_modifier = builder
                if builder == "latexpdf" or builder == "info":
                    flag = "-M"

                build_args = [flag, builder]
                target_build_dir = "{}/{}".format(data["outputdir"], builder_modifier)
                logger.debug(
                    "Running sphinx-build with args: %r",
                    build_args + current_argv,
                )
                # Create sphinx-build command
                cmd = (
                    sys.executable,
                    *get_python_flags(),
                    "-m",
                    "sphinx",
                    *build_args,
                    data["sourcedir"],
                    target_build_dir,
                    *current_argv,
                )
                current_cwd = os.path.join(data["basedir"], cwd_relative)
                env = os.environ.copy()
                env.update(
                    {
                        "SPHINX_MULTIVERSION_NAME": data["name"],
                        "SPHINX_MULTIVERSION_VERSION": data["version"],
                        "SPHINX_MULTIVERSION_RELEASE": data["release"],
                        "SPHINX_MULTIVERSION_SOURCEDIR": data["sourcedir"],
                        "SPHINX_MULTIVERSION_OUTPUTDIR": data["outputdir"],
                        "SPHINX_MULTIVERSION_CONFDIR": data["confdir"],
                    }
                )
                # Run sphinx-build
                subprocess.check_call(cmd, cwd=current_cwd, env=env)

                # Create artefacts if this build target should be downloadable
                if downloadable:
                    artefact_dir = "{}/artefacts".format(data["outputdir"])
                    os.makedirs(artefact_dir, exist_ok=True)
                    filename = "{project}_docs-{version}".format(
                        project=config.project.replace(" ", ""),
                        version=version_name.replace("/", "-"),
                    )

                    # Make an archive out of the build targets build directory
                    # Archive types supported by shutil.make_archive
                    if download_format in sphinx.ARCHIVE_TYPES:
                        shutil.make_archive(
                            "{}/{}-{}".format(
                                artefact_dir, filename, build_target_name
                            ),
                            download_format,
                            target_build_dir,
                        )
                    else:
                        # Find files matching project-name.extension, e.g.
                        # example.pdf in the target build directory
                        candidate_files = glob.glob(
                            "{build_dir}/**/*.{extension}".format(
                                build_dir=target_build_dir,
                                extension=download_format,
                            ),
                            recursive=True,
                        )
                        if len(candidate_files) > 1:
                            build_file_pattern = "{project}.{extension}".format(
                                project=config.project.replace(" ", ""),
                                extension=download_format,
                            )
                            build_artefacts = [
                                x
                                for x in candidate_files
                                if pathlib.Path(x.lower()).name
                                == build_file_pattern.lower()
                            ]
                        else:
                            build_artefacts = candidate_files
                        if len(build_artefacts) == 0:
                            logger.warning(
                                ("Build artefact {project}" "not found.").format(
                                    project=build_file_pattern.lower(),
                                )
                            )
                        elif len(build_artefacts) > 1:
                            logger.warning(
                                (
                                    "Ambiguous build artefact results: {}. "
                                    "Files not moved to artefact directory."
                                ).format(build_artefacts)
                            )
                        else:
                            # Found a single file with appropriate extension
                            shutil.copy(
                                build_artefacts[0],
                                os.path.join(
                                    artefact_dir,
                                    f"{filename}.{download_format}",
                                ),
                            )

                # Clean up build directory
                # Move html target to root of outputdir, if it exists.
                if build_target_name == "HTML":
                    shutil.copytree(
                        os.path.join(target_build_dir),
                        os.path.join(data["outputdir"]),
                        dirs_exist_ok=True,
                    )
                # Remove build directories
                if config.smv_clean_intermediate_files:
                    shutil.rmtree(target_build_dir)

            if config.smv_postbuild_command != "":
                logger.debug(
                    "Running postbuild command: %r",
                    config.smv_postbuild_command,
                )
                subprocess.check_call(
                    config.smv_postbuild_command, cwd=current_cwd, shell=True
                )
                if config.smv_postbuild_export_pattern != "":
                    matches = find_matching_files_and_dirs(
                        config.smv_postbuild_export_pattern, current_cwd
                    )
                    export_dst = os.path.join(
                        data["outputdir"],
                        config.smv_postbuild_export_destination,
                    )
                    export_to_destination(matches, export_dst)

        with resources.path(
            "exasol.sphinx_extensions.multiversion.templates", "multiversion-index.html"
        ) as p:
            with open(p) as f:
                template = Template(f.read())
            with open(os.path.join(args.outputdir, "index.html"), "w") as f:
                ref = gitrefs[-1]
                f.write(template.render(version=ref.name))

    return 0



import difflib
import io
from contextlib import ExitStack
from pathlib import Path
from typing import (
    Any,
    Mapping,
    Union,
)

import importlib_resources as resources
import typer
from rich.columns import Columns
from rich.console import Console
from rich.syntax import Syntax

stdout = Console()
stderr = Console(stderr=True)

CLI = typer.Typer()


def _workflows() -> Mapping[str, Any]:
    pkg = "exasol.toolbox.templates.github.workflows"

    def _normalize(name: str) -> str:
        name, ext = name.split(".")
        return name

    return {_normalize(w.name): w for w in resources.files(pkg).iterdir()}  # type: ignore


@CLI.command(name="list")
def list_workflows(
    columns: bool = typer.Option(
        False, "--columns", "-c", help="use column style presentation like `ls`"
    )
) -> None:
    """List all available workflows."""

    class List:
        def __init__(self, items: Any):
            self.items = items

        def __rich__(self) -> str:
            return "\n".join(self.items)

    output = List(_workflows()) if not columns else Columns(_workflows(), expand=True)
    stdout.print(output)


@CLI.command(name="show")
def show_workflow(
    workflow: str = typer.Argument(..., help="Workflow which shall be shown."),
) -> None:
    """Shows a specific workflow."""
    workflows = _workflows()
    if workflow not in workflows:
        stdout.print(f"Unknown workflow <{workflow}>.", style="red")
        raise typer.Exit(code=1)

    workflow = workflows[workflow]
    stdout.print(Syntax(workflow.read_text(), "yaml"))  # type: ignore


@CLI.command(name="diff")
def diff_workflow(
    workflow: str = typer.Argument(..., help="Workflow which shall be diffed."),
    dest: Path = typer.Argument(
        Path("./.github/workflows"),
        help="target directory to diff the workflow against.",
    ),
) -> None:
    """Diff a specific workflow against the installed one."""
    workflows = _workflows()
    if workflow not in workflows:
        stdout.print(f"Unknown workflow <{workflow}>.", style="red")
        raise typer.Exit(code=1)

    # Use Any type to enable reuse of the variable/binding name
    old: Any = dest / f"{workflow}.yml"
    new: Any = Path(_workflows()[workflow])
    with ExitStack() as stack:
        old = stack.enter_context(open(old) if old.exists() else io.StringIO(""))
        new = stack.enter_context(open(new))
        old = old.read().split("\n")
        new = new.read().split("\n")

    diff = difflib.unified_diff(old, new, fromfile="old", tofile="new")
    stdout.print(Syntax("\n".join(diff), "diff"))


def _install_workflow(
    src: Union[str, Path], dest: Union[str, Path], exists_ok: bool = False
) -> None:
    src, dest = Path(src), Path(dest)

    if dest.exists() and not exists_ok:
        raise FileExistsError("Workflow already exists")

    with ExitStack() as stack:
        input_file = stack.enter_context(open(src))
        output_file = stack.enter_context(open(dest, "w"))
        output_file.write(input_file.read())


def _select_workflows(workflow: str) -> Mapping[str, Any]:
    workflows = _workflows()
    if workflow != "all" and workflow not in workflows:
        raise Exception(f"Workflow <{workflow}> is unknown")
    workflows = (
        workflows
        if workflow == "all"
        else {name: path for name, path in workflows.items() if workflow == name}
    )
    return workflows


@CLI.command(name="install")
def install_workflow(
    workflow: str = typer.Argument("all", help="name of the workflow to install."),
    dest: Path = typer.Argument(
        Path("./.github/workflows"), help="target directory to install the workflow to."
    ),
) -> None:
    """
    Installs the requested workflow into the target directory.

    Attention: If there is an existing workflow with the same name it will be overwritten!
    """
    if not dest.exists():
        dest.mkdir()

    try:
        workflows = _select_workflows(workflow)
    except Exception as ex:
        stderr.print(f"[red]{ex}[/red]")
        raise typer.Exit(-1)

    for workflow, path in workflows.items():
        destination = dest / f"{workflow}.yml"
        _install_workflow(path, destination, exists_ok=True)
        stderr.print(f"Installed {workflow} in {destination}")


@CLI.command(name="update")
def update_workflow(
    workflow: str = typer.Argument("all", help="name of the workflow to install."),
    dest: Path = typer.Argument(
        Path("./.github/workflows"), help="target directory to install the workflow to."
    ),
    confirm: bool = typer.Option(
        False, help="Automatically confirm overwritting exsisting workflow(s)"
    ),
) -> None:
    """Similar to install but checks for existing workflows and shows diff"""
    if not dest.exists():
        dest.mkdir()

    try:
        workflows = _select_workflows(workflow)
    except Exception as ex:
        stderr.print(f"[red]{ex}[/red]")
        raise typer.Exit(-1)

    if confirm:
        install_workflow(workflow, dest)
        raise typer.Exit(0)

    for workflow, path in workflows.items():
        destination = dest / f"{workflow}.yml"
        try:
            _install_workflow(path, destination, exists_ok=False)
            stderr.print(f"Updated {workflow} in {destination}")
        except Exception:
            show_diff = typer.confirm(
                f"Workflow <{workflow}> already exists, show diff?"
            )
            if show_diff:
                diff_workflow(workflow, dest)

            overwrite = typer.confirm(f"Overwrite existing workflow?")
            if overwrite:
                _install_workflow(path, destination, exists_ok=True)
                stderr.print(f"Updated {workflow} in {destination}")


if __name__ == "__main__":
    CLI()



import nox

# imports all nox task provided by the toolbox
from exasol.toolbox.nox.tasks import *

# default actions to be run if nothing is explicitly specified with the -s option
nox.options.sessions = ["fix"]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import (
    Any,
    MutableMapping,
)

from nox import Session


@dataclass(frozen=True)
class Config:
    root: Path = Path(__file__).parent
    doc: Path = Path(__file__).parent / "doc"
    version_file: Path = Path(__file__).parent / "exasol" / "toolbox" / "version.py"

    @staticmethod
    def pre_integration_tests_hook(
        _session: Session, _config: Config, _context: MutableMapping[str, Any]
    ) -> bool:
        """Implement if project specific behaviour is required"""
        return True

    @staticmethod
    def post_integration_tests_hook(
        _session: Session, _config: Config, _context: MutableMapping[str, Any]
    ) -> bool:
        """Implement if project specific behaviour is required"""
        return True


PROJECT_CONFIG = Config()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['exasol',
 'exasol.toolbox',
 'exasol.toolbox.nox',
 'exasol.toolbox.pre_commit_hooks',
 'exasol.toolbox.sphinx',
 'exasol.toolbox.sphinx.multiversion',
 'exasol.toolbox.templates',
 'exasol.toolbox.tools']

package_data = \
{'': ['*'],
 'exasol.toolbox.sphinx.multiversion': ['templates/*'],
 'exasol.toolbox.templates': ['github/*',
                              'github/ISSUE_TEMPLATE/*',
                              'github/PULL_REQUEST_TEMPLATE/*',
                              'github/actions/python-environment/*',
                              'github/workflows/*']}

install_requires = \
['black>=23.1.0,<24.0.0',
 'coverage>=6.4.4,<8.0.0',
 'furo>=2022.9.15,<2023.0.0',
 'importlib-resources>=5.12.0,<6.0.0',
 'isort>=5.12.0,<6.0.0',
 'mypy>=0.971',
 'nox>=2022.8.7,<2023.0.0',
 'pre-commit>=3.1.1,<4.0.0',
 'prysk>=0.15.1,<0.16.0',
 'pylint>=2.15.4,<3.0.0',
 'pytest>=7.2.2,<8.0.0',
 'pyupgrade>=2.38.2,<4.0.0',
 'sphinx-copybutton>=0.5.0,<0.6.0',
 'sphinx>=5.3,<7.0',
 'typer[all]>=0.7.0']

entry_points = \
{'console_scripts': ['sphinx-multiversion = '
                     'exasol.toolbox.sphinx.multiversion.main:main',
                     'tbx = exasol.toolbox.tools.tbx:CLI',
                     'version-check = '
                     'exasol.toolbox.pre_commit_hooks.package_version:main']}

setup_kwargs = {
    'name': 'exasol-toolbox',
    'version': '0.6.1',
    'description': '',
    'long_description': 'Exasol Toolbox\n#####################\n\n.. image:: https://img.shields.io/pypi/v/exasol-toolbox\n     :target: https://pypi.org/project/exasol-toolbox/\n     :alt: PyPI Version\n\n.. image:: https://img.shields.io/pypi/pyversions/exasol-toolbox\n    :target: https://pypi.org/project/exasol-toolbox\n    :alt: PyPI - Python Version\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n    :alt: Formatter - Black\n\n.. image:: https://img.shields.io/badge/imports-isort-ef8336.svg\n    :target: https://pycqa.github.io/isort/\n    :alt: Formatter - Isort\n\n.. image:: https://img.shields.io/badge/typing-mypy-blue\n    :target: https://github.com/PyCQA/pylint\n    :alt: Pylint\n\n.. image:: https://img.shields.io/badge/pylint-7.8-green\n    :target: https://github.com/PyCQA/pylint\n    :alt: Pylint\n\n.. image:: https://img.shields.io/pypi/l/exasol-bucketfs\n     :target: https://opensource.org/licenses/MIT\n     :alt: License\n\n.. image:: https://img.shields.io/github/last-commit/nicoretti/python-toolbox\n     :target: https://pypi.org/project/exasol-toolbox/\n     :alt: Last Commit\n\n\n🚀 Features\n------------\n\n* Centrally managed standard tasks\n    - code formatting & upgrading\n    - linting\n    - type-checking\n    - unit-tests\n    - integration-tests\n    - coverage\n    - documentation\n\n* Centrally manged core workflows\n    - workspace/project verification\n    - build and publish releases\n    - build and publish documentation\n\n* Configurable & Extensible\n    - Project configuration\n    - Event hooks\n\n🔌️ Prerequisites\n-----------------\n\n- `Python <https://www.python.org/>`_ >= 3.8\n\n💾 Installation\n----------------\n\n.. code-block:: shell\n\n    pip install exasol-toolbox\n\n📚 Documentation\n----------------\n\nThe latest documentation can be found `here <https://exasol.github.io/python-toolbox/>`_\n',
    'author': 'Nicola Coretti',
    'author_email': 'nicola.coretti@exasol.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
